@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 1999-2001 Ricardo Ueda Karpischek
+  Copyright (C) 1999-2002 Ricardo Ueda Karpischek
 
   This is free software; you can redistribute it and/or modify
   it under the terms of the version 2 of the GNU General Public
@@ -2756,10 +2756,9 @@ int compose(int *l,int c)
         else
             return(1);
     }
-
     else if (*l == '\'') {
         if (c == '\'')
-            *l = '\'';
+            *l = '"';
         else if (c == 'a')
             *l = '';
         else if (c == 'e')
@@ -2860,12 +2859,14 @@ int compose(int *l,int c)
             *l = '';
         else if (c == 'U')
             *l = '';
+        else if (c == 's')
+            *l = '';
         else
             *l = c;
     }
     else if (*l == '`') {
         if (c == '`')
-            *l = '`';
+            *l = '"';
         else if (c == 'a')
             *l = '';
         else if (c == 'e')
@@ -2890,11 +2891,19 @@ int compose(int *l,int c)
             *l = c;
     }
     else if (*l == '/') {
-        if (c == '/')
+        if (c == '')
             *l = '/';
-        if (c == '?')
+        else if (c == '?')
             *l = '';
-        if (c == '!')
+        else if (c == '!')
+            *l = '';
+        else if (c == 'a')
+            *l = '';
+        else if (c == 'o')
+            *l = '';
+        else if (c == 'A')
+            *l = '';
+        else if (c == 'O')
             *l = '';
         else
             *l = c;
