@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 1999-2001 Ricardo Ueda Karpischek
+  Copyright (C) 1999-2002 Ricardo Ueda Karpischek
 
   This is free software; you can redistribute it and/or modify
   it under the terms of the version 2 of the GNU General Public
@@ -26,6 +26,7 @@ build.c: The function build.
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <math.h>
 #include "common.h"
 
 /*
@@ -105,10 +106,25 @@ void recog_validation(int i)
         "j", period, etc. An easy approximation for this criteria is
         considering the alignment. So we remove all SHAPE votes
         from dots with unknown alignment.
+
+        Obs. this rule is temporarily out of the code because it
+        drops a valid classification when the engine does not have
+        enough information to deduce the dot alignment.
     */
+    /*
     if ((m->tc == DOT) && (m->va < 0)) {
         rmvotes(SHAPE,k,-1,0);
     }
+    */
+
+    /*
+    *   COMMA with align = 01 => apostrophe. (GL)
+    */
+    if ((m->tc == COMMA) && (m->va == 01) 
+	&& (m->tr != NULL) && (strlen(m->tr->t) == 1)
+	&& (m->tr->t[0] == ',')) {
+        m->tr->t[0] = '\'';
+    }
 
     /*
         Apply add-hoc filtering rules.
@@ -146,10 +162,56 @@ void recog_validation(int i)
             db("oops.. untransliterated symbol classified as accent");
         }
 
-        else if ((b = bsymb(k,0)) >= 0) {
-            if (box_dist(ps[i],b,NULL) < ((mc[b].b-mc[b].t)/2)) {
-                for (t = &(mc[b].sl); (*t >= 0); t = &(mc[*t].sl));
-                *t = i;
+        else if (((b = bsymb(k,0)) >= 0) &&
+                 (box_dist(ps[i],b,NULL) < ((mc[b].b-mc[b].t)/2))) {
+
+            for (t = &(mc[b].sl); (*t >= 0); t = &(mc[*t].sl));
+            *t = i;
+            m->bs = i;
+        }
+
+        /*
+            double quote preliminar tests.
+        */
+        else if ((strcmp(trans(i),"\'")==0) &&
+                 ((b = lsymb(k,1)) >= 0) &&
+                 (strcmp(trans(b),"\'")==0)) {
+
+            for (t = &(mc[b].sl); (*t >= 0); t = &(mc[*t].sl));
+            *t = i;
+            m->bs = i;
+        }
+        else if ((strcmp(trans(i),"`") == 0) &&
+                 ((b = lsymb(k,1)) >= 0) &&
+                 (strcmp(trans(b),"`") == 0)) {
+
+            for (t = &(mc[b].sl); (*t >= 0); t = &(mc[*t].sl));
+            *t = i;
+            m->bs = i;
+        }
+    }
+
+    /* attach dot to dot/comma -> colon/semicolon (GL)*/
+    if (m->tc == DOT) {
+        int *t;
+
+        /*
+           strangely enough, it seems that some untransliterated symbols
+           are being classified as dots. (??? Copied from accents. GL)
+        */
+        if (m->tr == NULL) {
+            db("oops.. untransliterated symbol classified as dot");
+        }
+
+        else if ( (b = bsymb(k,1)) >= 0 ) {
+	         /* printf("\n<%d on %d>",i,b); */
+            if ( ((mc[b].tc == DOT) || (mc[b].tc == COMMA))  
+		 && (mc[b].l < mc[i].r) && (mc[i].l < mc[b].r) 
+	       ) {
+	         /* printf("!%d-%d ; %d-%d",mc[i].l,mc[i].r,mc[b].l,mc[b].r); */
+                 for (t = &(mc[b].sl); (*t >= 0); t = &(mc[*t].sl));
+                 *t = i;
+                 m->bs = i;
             }
         }
     }
@@ -157,19 +219,236 @@ void recog_validation(int i)
 
 /*
 
-Returns -1 if the line a precedes the line b,
-or +1 if the line b precedes the line a, or
-0 if cannot decide.
+Using the estimated baseline for each symbol, computes the baseline
+geometry, top and bottom limits for line a.
+
+          +-+        +-+
+       +-+| |  +-++-+| |+-++-+
+    a: | || |  | || || || || |
+       +*++*+  |*|O*++*++*++*+
+               +-+
+
+For each symbol s on line a, estimates its baseline bl=bl(s). The
+medium points at heigh bl(s) for each symbol s (stars, in the
+figure) are used to interpolate a line. The slope of this line is
+returned as *sl. The topmost and bottommost coordinates reached
+by the line a symbols are returned on *at and *ab. The baseline
+medium point ("O" in the figure) is returned on (*mx,*my). The
+medium x_height is returned in *xh. The function value is 0 if
+could not interpolate a baseline, 1 otherwise.
+
+*/
+int bl_geo(int a,float *sl,int *at,int *ab,int *mh,float *mx,float *my)
+{
+    /* alignment */
+    int as,xh,bl,ds;
+
+    /* liner regression */
+    int x,y,n;
+    int sxh,nxh;
+    long long sx,sy,sx2,sy2,sxy;
+    float N;
+
+    /* other */
+    int s,w,ret,l,r;
+
+    /* prepare */
+    *at = YRES;
+    *ab = 0;
+    l = XRES;
+    r = 0;
+    sx = sy = sx2 = sy2 = sxy = 0;
+    sxh = 0;
+    n = nxh = 0;
+
+    /* visit all symbols */
+    for (w=line[a].f; w>=0; w=word[w].E) {
+        for (s=word[w].F; s>=0; s=mc[s].E) {
+
+            /* extreme coordinates */
+            if (mc[s].b > *ab)
+                *ab = mc[s].b;
+            if (mc[s].t < *at)
+                *at = mc[s].t;
+            if (mc[s].l < l)
+                l = mc[s].l;
+            if (mc[s].r > r)
+                r = mc[s].r;
+
+            /* alignment parameters for a and b */
+            get_ap(s,&as,&xh,&bl,&ds);
+
+            /* accumulate */
+            if (bl > 0) {
+                x = (mc[s].l + mc[s].r) / 2;
+                y = bl;
+                sx  += x;
+                sx2 += (x*x);
+                sy  += y;
+                sy2 += (y*y);
+                sxy += (x*y);
+                ++n;
+            }
+
+            /* accumulate xh */
+            if (xh > 0) {
+                sxh += xh;
+                ++nxh;
+            }
+
+        }
+    }
+
+    /* slope and medium point */
+    if ((n > 1) && (nxh > 0)) {
+        float y0;
+
+        N = n;
+        *sl = (sxy - sx*sy/N) / (sx2 - sx*sx/N);
+        *mx = ((float) (l+r)) / 2;
+
+        /*
+            If the absolute value of the slope is too large, our
+            dataset is strange and probably useless.
+        */
+        if (fabs(*sl) > 0.5) {
+            db("what? slope %f?\n",*sl);
+            ret = 0;
+        }
+
+        /* finish computing the medium point */
+        else {
+            y0 = (sy - *sl*sx) / N;
+            *my = *sl * *mx + y0;
+            *mh = (sxh / nxh);
+            ret = 1;
+        }
+    }
+
+    else {
+        ret = 0;
+    }
+
+    return(ret);
+}
+
+/*
+
+Returns -1 if the line a precedes the line b, or +1 if the line b
+precedes the line a, or 0 if cannot decide.
+
+This is the service used to sort the lines when generating the
+OCR output. Clara OCR uses the alignment heuristics to build
+words from the symbols, and build lines from the words. The
+alignment heuristics do not deduce if one line a precedes one
+line b, this is done by a subsequent build step.
+
+       +---------------+
+    a: |to be or not to|
+       +---------------+
+
+       +-----------------------+
+    b: |be, that's the question|
+       +-----------------------+
+
+Due to skew, it's not that easy to deduce if one line precedes
+another. Remember that not all lines are full lines. Some lines
+are only partially filled. Also, some lines become broken due to
+both due to weaknesses on the current alignment
+heuristics.
 
 */
 int cmpln(int a,int b)
 {
     sdesc *p,*q;
+    int at,ab,ah,bt,bb,bh;
+    int ag,bg;
+    float as,bs,ax,ay,bx,by;
 
+    /* get the first symbol on each line */
     p = mc + word[line[a].f].F;
     q = mc + word[line[b].f].F;
 
-    if ((p->b+p->t)/2 + 15 < (q->b+q->t)/2)
+    /*
+        When the lines belong to different zones, use
+        the zone indexes to decide.
+    */
+    if (p->c < q->c)
+        return(-1);
+    else if (p->c > q->c)
+        return(1);
+
+    /* geometries of the baselines */
+    ag = bl_geo(a,&as,&at,&ab,&ah,&ax,&ay);
+    bg = bl_geo(b,&bs,&bt,&bb,&bh,&bx,&by);
+
+    /*
+
+        Analysis of the baselines
+        -------------------------
+
+        At this point, we interpolated the baselines for
+        lines a and b, and computed their slopes. Unless
+        the computation failed, now we are going to study
+        the relative position of the baselines.
+
+
+                    Q
+              ......*..........._____________ (a)
+                    |
+                    |
+              ------*------- (b)
+                    M
+
+        We proceed as follows: istead of the baseline of a, we'll
+        consider the medium slope (between the two baselines) and
+        draw the line with such slope and intercepting the medium
+        point of the baseline of a. Now we consider the medium
+        point M of the baseline of b and its vertical projection
+        Q into that line. The relative position of M and Q is then
+        studied.
+    */
+    if (ag && bg) {
+        float s,A,B,C,y,a,b;
+
+        /* medium slope */
+        a = atan(as);
+        b = atan(bs);
+        if (as > bs) {
+            s = a - (a-b)/2;
+        }
+        else {
+            s = b - (b-a)/2;
+        }
+
+        /* equation of the baseline for line a (normal form Ax+By+C=0) */
+        A = s;
+        B = -1;
+        C = -(A*ax+B*ay);
+
+        /*
+            vertical coordinate of Q.
+        */
+        y = C + A*ax;
+
+        /* the vertical distance is large enough: decide */
+        if (fabs(y-by) > ((ah+bh)/2)) {
+
+            if (y < by)
+                return(-1);
+            else
+                return(1);
+        }
+    }
+
+    /*
+        Try some simple criterion (note the hardcoded values).
+    */
+    if (ab < bt)
+        return(-1);
+    else if (bb < at)
+        return(1);
+    else if ((p->b+p->t)/2 + 15 < (q->b+q->t)/2)
         return(-1);
     else if ((q->b+q->t)/2 + 15 < (p->b+p->t)/2)
         return(1);
@@ -353,6 +632,7 @@ w_pair tests word pairing, and returns the following diagnostics:
 1 .. insufficient vertical intersection
 2 .. incomplete data
 3 .. maximum horizontal distance exceeded
+4 .. different zones
 
 */
 int w_pair(int wa,int wb)
@@ -371,6 +651,9 @@ int w_pair(int wa,int wb)
         return(2);
     }
 
+    if (mc[a].c != mc[b].c)
+        return(4);
+
     /* compute horizontal distance */
     if (mc[a].r < mc[b].l)
         d = mc[b].l - mc[a].r;
@@ -479,11 +762,13 @@ void diag_wpairing(int t)
         show_hint(2,"incomplete data");
     if (r == 3)
         show_hint(2,"maximum horizontal distance exceeded");
+    if (r == 4)
+        show_hint(2,"different zones");
 }
 
 /*
 
-Diagnose pairing against current symbol.
+Diagnose symbol pairing.
 
 */
 void diag_pairing(int t)
@@ -502,16 +787,18 @@ void diag_pairing(int t)
 
     if (r == 0)
         show_hint(2,"pairing successful");
-    if (r == 1)
+    else if (r == 1)
         show_hint(2,"insuficcient vertical intersection");
-    if (r == 2)
+    else if (r == 2)
         show_hint(2,"one or both symbols above ascent");
-    if (r == 3)
+    else if (r == 3)
         show_hint(2,"one or both symbols below descent");
-    if (r == 4)
+    else if (r == 4)
         show_hint(2,"maximum horizontal distance exceeded");
-    if (r == 5)
+    else if (r == 5)
         show_hint(2,"incomplete data");
+    else if (r == 6)
+        show_hint(2,"different zones");
 }
 
 /*
@@ -639,7 +926,7 @@ int build(int reset)
 
         */
         for (m=0; m<=tops; ++m) {
-            mc[m].E = mc[m].W = mc[m].sw = mc[m].sl = -1;
+            mc[m].E = mc[m].W = mc[m].sw = mc[m].sl = mc[m].bs = -1;
         }
 
         /* (devel)
@@ -674,9 +961,8 @@ int build(int reset)
             2. Distributing symbols on words
 
             The second step is, for each CHAR not in any word, define
-            a unitary word around it and extend it to right
-            and left applying the symbol pairing test. When
-            extending, merge words when necessary.
+            a unitary word around it, and extend it to right
+            and left applying the symbol pairing test.
 
         */
         if (m <= topps) {
@@ -944,18 +1230,37 @@ int build(int reset)
 
         /*
             Build unitary words for baseline-aligned dots and commas
+            and other things.
         */
         for (i=0; i<=topps; ++i) {
 
             m = mc + (k = ps[i]);
 
-            if ((m->sw < 0) &&
-                ((m->tc == DOT) || (m->tc == COMMA))) {
-
-                if ((m->va == 22) || (m->va == 23)) {
-                    new_word(k);
-                }
-            }
+            if (
+		((m->sw < 0) &&
+                 (m->tc == DOT) &&
+                 (m->va == 22)) 
+		||
+	        ((m->sw < 0) &&
+                 (m->tc == COMMA) &&
+                 ((m->va == 1) /* apostrophe. (GL) */
+		  ||(m->va == 23)))
+                ||
+	        ((m->sw < 0) &&
+                 (m->tc == ACCENT) &&
+                 (m->bs < 0) &&
+                 (strcmp(trans(k),"'")==0) &&
+                 (m->va == 1))
+                ||
+	        ((m->sw < 0) &&
+                 (m->tc == ACCENT) &&
+                 (m->bs < 0) &&
+                 (strcmp(trans(k),"`")==0) &&
+                 (m->va == 1))
+	       )
+	       {
+                   new_word(k);
+               }
         }
 
         /* prepare next state */
