@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 1999-2001 Ricardo Ueda Karpischek
+  Copyright (C) 1999-2002 Ricardo Ueda Karpischek
 
   This is free software; you can redistribute it and/or modify
   it under the terms of the version 2 of the GNU General Public
@@ -83,8 +83,17 @@ void toasc(FILE *F,char *name,int type,int n,void *f)
         n = strlen((char *)f);
     }
 
+    /* empty array */
+    if ((n==0) && (type != TNULL)) {
+        r = snprintf(s+d,max-d,"=");
+        if (r < 0) {
+            fatal(BO,"buffer overflow at toasc\n");
+        }
+        d += r;
+    }
+
     /* dump array */
-    for (sep='=', i=0; i<n; ++i) {
+    else for (sep='=', i=0; i<n; ++i) {
 
         if (d > max-128) {
             if (F == NULL)
@@ -122,7 +131,7 @@ void toasc(FILE *F,char *name,int type,int n,void *f)
     if (F == NULL)
         totext("%s",s);
     else
-        fprintf(F,"%s",s);
+        zfwrite(s,strlen(s),1,F);
 }
 
 /*
@@ -393,6 +402,7 @@ int dump_session(char *f,int reset)
     if (st == 7) {
         fprintf(F,"%s","</Clara>\n");
         zfclose(F,pio);
+        sess_changed = 0;
         return(0);
     }
 
@@ -412,8 +422,9 @@ int dump_session(char *f,int reset)
         toasc(F,"<unstruct",TNULL,0,NULL);
         toasc(F,"XRES",TINT,1,&XRES);
         toasc(F,"YRES",TINT,1,&YRES);
-        toasc(F,"limits_top",TINT,1,&limits_top);
-        toasc(F,"limits",TINT,8,limits);
+        toasc(F,"zones",TINT,1,&zones);
+        if (zones > 0)
+            toasc(F,"limits",TINT,8*zones,limits);
         toasc(F,"curr_mc",TINT,1,&curr_mc);
         toasc(F,"symbols",TINT,1,&symbols);
         toasc(F,"doubts",TINT,1,&doubts);
@@ -489,6 +500,7 @@ int dump_session(char *f,int reset)
             toasc(F,"W",TINT,1,&(mc[i].W));
             toasc(F,"sw",TINT,1,&(mc[i].sw));
             toasc(F,"sl",TINT,1,&(mc[i].sl));
+            toasc(F,"bs",TINT,1,&(mc[i].bs));
             toasc(F,"tc",TCHAR,1,&(mc[i].tc));
             toasc(F,"pb",TINT,1,&(mc[i].pb));
             toasc(F,"lfa",TINT,1,&(mc[i].lfa));
@@ -776,6 +788,7 @@ int dump_patterns(char *f,int reset)
         toasc(F,"BM",TFLOAT,1,&BM);
         toasc(F,"PG_AUTO",TINT,1,&PG_AUTO);
         toasc(F,"TC",TINT,1,&TC);
+        toasc(F,"SL",TINT,1,&SL);
 
         /* magic numbers */
         toasc(F,"m_mwd",TINT,1,&m_mwd);
@@ -917,8 +930,9 @@ int recover_session(char *f,int st,int reset)
     else if (fromasc(F,"<unstruct",TNULL,0,NULL) == 0) {
         fromasc(F,"XRES",TINT,1,&XRES);
         fromasc(F,"YRES",TINT,1,&YRES);
-        fromasc(F,"limits_top",TINT,1,&limits_top);
-        fromasc(F,"limits",TINT,8,limits);
+        fromasc(F,"zones",TINT,1,&zones);
+        if (zones > 0)
+            fromasc(F,"limits",TINT,8*zones,limits);
         fromasc(F,"curr_mc",TINT,1,&curr_mc);
         fromasc(F,"symbols",TINT,1,&symbols);
         fromasc(F,"doubts",TINT,1,&doubts);
@@ -974,6 +988,8 @@ int recover_session(char *f,int st,int reset)
         fromasc(F,"W",TINT,1,&(m.W));
         fromasc(F,"sw",TINT,1,&(m.sw));
         fromasc(F,"sl",TINT,1,&(m.sl));
+        m.bs = -1;
+        fromasc(F,"bs",TINT,1,&(m.bs));
         fromasc(F,"tc",TCHAR,1,&(m.tc));
         fromasc(F,"pb",TINT,1,&(m.pb));
         fromasc(F,"lfa",TINT,1,&(m.lfa));
@@ -1160,8 +1176,9 @@ int recover_session(char *f,int st,int reset)
             }
         }
 
-        /* finish */
+        /* finished */
         zfclose(F,pio);
+        sess_changed = 0;
         return(0);
     }
 
@@ -1190,7 +1207,7 @@ int recover_acts(char *f)
     else {
         if ((F = zfopen(f,"r",&pio)) == NULL) {
             if (verbose)
-                warn("acts file \"%s\" notfound (ok)",f);
+                warn("acts file \"%s\" not found (ok)",f);
             return(0);
         }
     }
@@ -1271,8 +1288,10 @@ int recover_acts(char *f)
         else {
             if (f != NULL)
                 fatal(FD,"unexpected token at line %d of file %s",DLINE,f);
-            else
+            else {
+                printf(text);
                 fatal(FD,"unexpected token found on internally generated act");
+            }
         }
     }
 
@@ -1336,6 +1355,8 @@ int recover_patterns(char *f)
             fromasc(F,"BM",TFLOAT,1,&BM);
             fromasc(F,"PG_AUTO",TINT,1,&PG_AUTO);
             fromasc(F,"TC",TINT,1,&TC);
+            SL = 0;
+            fromasc(F,"SL",TINT,1,&SL);
 
             /* magic numbers */
             fromasc(F,"m_mwd",TINT,1,&m_mwd);
@@ -1537,7 +1558,7 @@ int free_page(void)
 
 /*
 
-From the page name recover its number.
+From the page name, recover its number.
 
 */
 int pagenbr(char *p)
@@ -1560,20 +1581,32 @@ int pagenbr(char *p)
 
 Prepare the names to load the current doc.
 
+Clara OCR maintains prepared various different names associated with
+the page name. For instance, when the file "1.pgm" is loaded, the
+variable "session" contains the string "1.session". All such strings
+are prepared here in order to simplify opening files.
+
 */
 void names_cpage(void)
 {
     int i,j,n;
 
+    /* locate the position at pagelist of the page cpage */
     for (n=j=0; n<cpage; ++n)
         while (pagelist[j++] != 0);
+
+    /* page name (e.g. "1.pgm") */
     strcpy(pagename,pagelist+j);
+
+    /* page base ("1", in our example) */
     strcpy(pagebase,pagename);
     i = strlen(pagebase);
-    if ((i>4) && (strcmp(pagebase+i-4,".pbm")==0))
+    if ((i>4) && ((strcmp(pagebase+i-4,".pbm")==0) || (strcmp(pagebase+i-4,".pgm")==0)))
         pagebase[i-4] = 0;
-    if ((i>7) && (strcmp(pagebase+i-7,".pbm.gz")==0))
+    if ((i>7) && ((strcmp(pagebase+i-7,".pbm.gz")==0) || (strcmp(pagebase+i-7,".pgm.gz")==0)))
         pagebase[i-7] = 0;
+
+    /* session (workdir path plus "1.session" in our example) */
     strcpy(session,workdir);
     strcat(session,pagebase);
     if (zsession)
@@ -1587,19 +1620,20 @@ void names_cpage(void)
 Load specified page.
 
 */
-int load_page(int p,int reset)
+int load_page(int p,int reset,int bin)
 {
     int x0,y0;
-    static int e;
+    static int mode;
+    char f[MAXFNL+1];
 
+    /* prepare */
     if (reset) {
-        char f[MAXFNL+1];
 
         /* no need to read the current page */
-        if (p == cpage)
+        if ((p == cpage) && (pixmap != NULL) && (!bin))
             return(0);
 
-        /* set new current page */
+        /* set the new current page */
         cpage = p;
         names_cpage();
 
@@ -1611,47 +1645,101 @@ int load_page(int p,int reset)
         doubts = 0;
         words = 0;
 
-        /*
-            Start reading the session file, if we have one,
-            or the bitmap.
-        */
-        if ((e = recover_session(session,0,1)) == 0) {
+        /* Try to start recovering the session file */
+        if (recover_session(session,0,1)) {
+            mode = 1;
+        }
+
+        /* Try to start reading the bitmap */
+        else {
+            int t;
+
+            /* filename */
             strncpy(f,pagesdir,MAXFNL);
             strncat(f,pagename,MAXFNL);
-            pbm2bm(f,1);
+            t = strlen(f);
+
+            /* start binarization/segmentation */
+            if ((pixmap != NULL) && (bin)) {
+
+                /* use the local thresholder */
+                if (localbin_strong) {
+                    find_thing(pixmap,1,-1,-1);
+                    mode = 2;
+                }
+
+                /* use the global thresholder */
+                else {
+                    sh_tries = 0;
+                    pbm2bm(pixmap,1);
+                    mode = 2;
+                }
+            }
+
+            /* start reading pixmap from file */
+            else {
+                loadpgm(1,f,&pixmap,&XRES,&YRES);
+                mode = 3;
+            }
         }
-        cl_ready = 0;
+
         return(1);
     }
 
-    if (e) {
-        e = recover_session(NULL,0,0);
-        if (e)
+    /* continue recovering the session file */
+    if (mode == 1) {
+
+        /* not finished yet */
+        if (recover_session(NULL,0,0))
             return(1);
+
+        /* finished reading the session file */
         cl_ready = 1;
         synchronize();
     }
 
-    if (cl_ready == 0) {
+    /* continue reading the bitmap */
+    else if (mode == 2) {
 
-        /* read the bitmap and compute the closures */
-        if (pbm2bm(NULL,0))
+        /* not finished yet */
+        if (localbin_strong) {
+            if (find_thing(NULL,0,-1,-1))
+                return(1);
+        }
+
+        /* global thresholding */
+        else {
+            if (pbm2bm(NULL,0))
+                return(1);
+        }
+
+        /* finished reading the bitmap */
+        cl_ready = 1;
+    }
+
+    /* continue reading the pixmap */
+    else {
+
+        /* continue reading */
+        if (loadpgm(0,f,&pixmap,&XRES,&YRES))
             return(1);
 
-        cl_ready = 1;
+        /* PAGE only is mandatory for graymaps */
+        if (pm_t == 8)
+            *cm_o_pgo = 'X';
 
-        /* predefined zone */
-        limits_top = 4;
-        limits[0] = 0;
-        limits[1] = 0;
-        limits[2] = XRES-1;
-        limits[3] = YRES-1;
-        limits[4] = 0;
-        limits[5] = YRES-1;
-        limits[6] = XRES-1;
-        limits[7] = 0;
+        /* no predefined zone */
+        zones = 0;
+
+        /* finished */
+        sess_changed = 0;
+        return(0);
     }
 
+    /* set the window title */
+    if (batch_mode == 0)
+        swn(pagebase);
+
     /* list the preferred symbols */
     make_pmc();
 
@@ -1673,10 +1761,6 @@ int load_page(int p,int reset)
     dw[PAGE_LIST].x0 = y0;
     dw[PAGE_LIST].y0 = y0;
 
-    /* set the window title */
-    if (batch_mode == 0)
-        swn(pagebase);
-
     /* prepare data structures for list_cl */
     list_cl(0,0,0,0,1);
 
@@ -1690,5 +1774,6 @@ int load_page(int p,int reset)
     dl_c[p] = classes;
 
     /* finished */
+    sess_changed = 0;
     return(0);
 }
