@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 1999-2001 Ricardo Ueda Karpischek
+  Copyright (C) 1999-2002 Ricardo Ueda Karpischek
 
   This is free software; you can redistribute it and/or modify
   it under the terms of the version 2 of the GNU General Public
@@ -31,6 +31,13 @@ extern char *version;
 
 /*
 
+Epsilon.
+
+*/
+extern float eps;
+
+/*
+
 Alphabets.
 
 Obs. due to the pattern types accounting features, alphabets IDs
@@ -181,18 +188,20 @@ step to the
 */
 #define OCR_SAVE 1
 #define OCR_LOAD 2
-#define OCR_CONS 3
-#define OCR_OPT 4
-#define OCR_REV 5
-#define OCR_BLOCK 6
-#define OCR_CLASS 7
-#define OCR_GEOMERGE 8
-#define OCR_BUILD 9
-#define OCR_SPELL 10
-#define OCR_OUTP 11
-#define OCR_DOUBTS 12
-#define OCR_OTHER 13
-#define LAST_STEP 13
+#define OCR_PREPROC 3
+#define OCR_BLOCK 4
+#define OCR_SEG 5
+#define OCR_CONS 6
+#define OCR_OPT 7
+#define OCR_REV 8
+#define OCR_CLASS 9
+#define OCR_GEOMERGE 10
+#define OCR_BUILD 11
+#define OCR_SPELL 12
+#define OCR_OUTP 13
+#define OCR_DOUBTS 14
+#define OCR_OTHER 15
+#define LAST_STEP 15
 
 /*
 
@@ -204,6 +213,7 @@ Revision operations.
 #define REV_SLINK 3
 #define REV_ALINK 4
 #define REV_DIS 5
+#define REV_PATT 6
 
 /*
 
@@ -235,6 +245,28 @@ LINKMUL*FS is the total width that an horizontal link may have.
 #define LINKMUL 3
 #define LFS (LINKMUL*FS)
 
+/*
+
+The pixmap
+
+*/
+extern unsigned char *pixmap;
+extern int thresh_val,pm_t;
+extern int graydist[];
+extern int pp_deskew,pp_bal,pp_double,pp_deskew_accurate;
+extern float pp_dark, pp_avoid_links;
+extern int p_match;
+extern int bin_method;
+
+/*
+
+Binarization methods
+
+*/
+#define gl_thresh (bin_method == 2)
+#define localbin ((bin_method == 3) || (bin_method == 4))
+#define localbin_strong (bin_method == 4)
+
 /* (devel)
 
 Global variables
@@ -367,9 +399,9 @@ Symbol transliterations
 
 Clara OCR maintains a list of 0 or more proposed or deduced
 transliterations for each symbol. Along the OCR process, each
-transliteration receives "votes" from reviewers or from machine
-deduction heuristics, based on shape similarity or on
-spelling.
+transliteration receives "votes" from reviewers (REVISION votes)
+or from machine deduction heuristics, based on shape similarity
+(SHAPE votes) or on spelling (SPELLING votes).
 
 So the choice of the "best" transliteration is performed through
 election. Votes are stored on structures of type vdesc, and
@@ -378,10 +410,27 @@ symbol stores a pointer for a (possibly empty) list of
 transliterations and each transliteration stores a pointer
 for a (possibly empty) list of votes.
 
-As the total stored information about one symbol may be large,
-Clara maintains for each symbol its "transliteration class", used
-by the heuristics to categorize each symbol and also to test the
-current transliteration status (is it known? is it dubious?).
+So, for instance, when one classifier deduces that one symbol is
+"a", it generates a "vote" for the transliteration of that symbol
+to be "a". At the same time, another heuristic could generate
+another vote for the transliteration to be, say, "o". The diagram
+illustrates this situation:
+
+   sdesc  ---> trdesc ("a")  ---> trdesc ("o")
+                 |                  |
+                 +-vdesc            + vdesc
+                 |
+                 +-vdesc
+
+In this case, the transliteration "a" has two votes, one from the
+classifier and another from, say, revision and the transliteration "o"
+has one vote.
+
+As the total stored information about one symbol may be large, Clara
+maintains for each symbol its "transliteration class", used by the
+heuristics to categorize each symbol and also to test the current
+transliteration status (is it known? is it dubious?), frequently used
+along the source code.
 
 */
 typedef struct {
@@ -645,8 +694,8 @@ typedef struct  {
     /* geometric data */
     int l,r,t,b;     /* geometric limits */
     int nbp;         /* number of black pixels */
-    char va;         /* probable vertical alignment */
-    char c;          /* text column */
+    char va;         /* vertical alignment */
+    char c;          /* column (zone) */
 
     /*
         Flags. Symbols have per-transliteration flags
@@ -667,6 +716,7 @@ typedef struct  {
     int N,S,E,W;     /* nearest symbols */
     int sw;          /* surrounding word */
     int sl;          /* list of signals (accents) */
+    int bs;          /* base symbol */
 
     /* transliterations */
     trdesc *tr;      /* first transliteration */
@@ -680,6 +730,27 @@ typedef struct  {
 
 /*
 
+Symbol transliteration shortcut.
+
+This macro returns always a non-null pointer. If the symbol is
+untransliterated, it returns a pointer to an empty string (""). Be
+careful! this is for immediate usage only. Example:
+
+    if (strcmp(trans(m),".") == 0)
+
+Never use as follows:
+
+    char *p;
+
+    p = trans(m);
+    (...)
+    if (strcmp(p,".") == 0)
+
+*/
+#define trans(b) (((mc[b].tr==NULL) || ((mc[b].tr)->t==NULL)) ? "" : ((mc[b].tr)->t))
+
+/*
+
 General statuses, that apply for various structure fields.
 
 */
@@ -730,6 +801,7 @@ extern int shape_opt;
 extern int square_opt;
 extern int copy_opt;
 extern int i64_opt;
+extern int disp_opt;
 
 /* from skel.c */
 extern int   SA;
@@ -775,12 +847,9 @@ extern int XRES,YRES,DENSITY;
 Common buffer for HTML generation
 
 */
-extern char *text;
+extern char *text,*ptext;
 extern int textsz,topt;
-
-/* OCR zone */
-extern int limits[];
-extern int limits_top;
+extern int ptextsz,ptopt;
 
 /*
 
@@ -843,10 +912,10 @@ typedef struct {
     short ds;          /* size of the field d */
     char *tr;          /* transliteration */
     char *d;           /* name of the originating page */
-    int e;             /* originating mc */
+    int e;             /* originating symbol (DEPRECATED) */
     float sl;          /* slope (skew) */
     short l,t;         /* fragment top-left position within symbol */
-    short p[8];        /* parameters for skeleton computation */
+    short p[8];        /* parameters for skeleton computation (deprecated?) */
 } pdesc;
 
 /*
@@ -943,6 +1012,7 @@ flag:
 #define F_ISNULL 256
 #define F_SDIM 512
 #define F_BAD 1024
+#define F_ATTACH 2048
 
 /*
 
@@ -969,9 +1039,6 @@ extern pdesc *pattern;
 extern int s_id;
 extern int *slist,slistsz;
 
-/* are closures ready? */
-extern int cl_ready;
-
 /*
 
 OCR statuses.
@@ -1008,7 +1075,7 @@ extern float LW,
              TM,
              BM;
 
-extern int PG_AUTO, TC;
+extern int PG_AUTO, TC, SL;
 
 /*
 
@@ -1039,15 +1106,29 @@ Session flags.
 
 */
 extern int finish;
+extern int sess_changed;
 extern int font_changed;
 extern int hist_changed;
 extern int form_changed;
+extern int cl_ready;
+extern int tskel_ready;
 extern int wnd_ready;
 extern int web;
 extern int batch_mode;
 extern int nullifying;
 extern int dkeys;
 extern int zsession;
+extern int dmmode;
+
+/*
+
+Type of action to perform, when the user trains an already
+classified symbol.
+
+*/
+extern int action_type;
+extern char *r_tr_old,*r_tr_new;
+extern int r_tr_patt,r_tr_symb;
 
 /*
 
@@ -1069,6 +1150,7 @@ TUNE form.
 */
 extern int classifier;
 extern int bf_auto,st_auto;
+extern float st_bq;
 extern int mf_i,mf_b,mf_c,mf_r,mf_l;
 
 /* override color buttons */
@@ -1173,7 +1255,8 @@ indexes and must be defined as 0, 1 and 2.
 #define TUNE_ACTS 3
 #define TUNE_SKEL 13
 #define PATTERN_TYPES 14
-#define TOP_DW 14
+#define PATTERN_ACTION 15
+#define TOP_DW 15
 
 /* maximum size of revision record (without LF) */
 #define MAXRRL 240
@@ -1204,28 +1287,25 @@ How to add an application button
 
 These are the steps to add a new button:
 
-1. Increase the definition of BUTT_PER_COL by 1, and create a new
-button macro after those already existing (bzoom, balpha,
-etc). Note that each button macro is defined as an unique index
-on the range 0..BUTT_PER_COL-1 (any permutation is ok).
+1. Create a new button macro after those already existing (bzoom,
+balpha, etc). Note that each button macro is defined as an unique
+integer (0, 1, 2, etc).
 
-  #define BUTT_PER_COL 14
+  #define bzoom 0
   [...]
   #define bfoo 13
 
-2. Define the label for this button (just add the corresponding
-entry to the initialization block of the array BL). Multi-state
-buttons have multiple labels, specified as
+2. Register the new button at init_ds(), together with its
+label. Multi-state buttons have multiple labels, specified as
 "state1:state2:state3":
 
-  char *BL[] = {
-      [...]
-      "foo"
-  }
+    register_button(bzoom,"zoom");
+    [...]
+    register_button(bfoo,"foo");
 
-The current state of the button is stored by button[bfoo]. When
-the state is nonzero, the button is drawn using a dark
-background.
+The current state of the new button is stored by
+button[bfoo]. When the state is nonzero, the button is drawn
+using a dark background.
 
 3. Add a new block to attend this button on mactions_b1 and, if
 desired, on mactions_b2 (just copy one existing block and adapt
@@ -1251,20 +1331,14 @@ the button state, it must request redraw. Example:
       redraw_button = bfoo;
 
 */
-#define BUTT_PER_COL 9
-extern char button[];
-extern int dmmode;
+extern int BUTT_PER_COL,bl_sz;
+extern char *button,**BL;
+extern int nalpha;
+
 
 /*
 
-The operation modes that determine (in part) the
-behaviour of the mouse button 1.
-
-*/
-
-/*
-
-other buttons
+GUI buttons
 
 */
 #define bzoom 0
@@ -1273,9 +1347,8 @@ other buttons
 #define bstop 3
 #define balpha 4
 #define btype 5
-#define bbold 6
-#define bitalic 7
-#define bbad 8
+#define bbad 6
+#define btest 7
 
 /*
 
@@ -1312,13 +1385,24 @@ Types of menu entries:
     use CM_TA for actions (performed by cma).
     use CM_TR for multiple choice.
 
+Per-item flags:
+
+    CM_NG this item starts a new group
+    CM_H  this item is (currently) hidden
+    CM_D  dismiss the menu when this item is selected
+
+MAX_MT is the maximum number of characters accepted for menu titles
+and item labels.
+
 */
-#define CM_TB 1
-#define CM_TA 2
-#define CM_TR 3
-#define CM_TM 7
-#define CM_NG 8
-#define MAX_MT 35
+#define CM_TB  1
+#define CM_TA  2
+#define CM_TR  3
+#define CM_TM  7
+#define CM_NG  8
+#define CM_HI 16
+#define CM_DI 32
+#define MAX_MT 45
 
 /*
 
@@ -1330,7 +1414,9 @@ typedef struct {
     int m;              /* maximum number of labels (may grow) */
     int n;              /* effective number of labels */
     char *l;            /* labels */
-    char *t;            /* item types */
+    char *t;            /* per-item types and flags */
+    char **h;           /* per-item short help */
+    int *f;             /* per-item activation flags */
     int p;              /* position on bar */
     int c;              /* current item */
     char tt[MAX_MT+1];  /* menu title */
@@ -1347,27 +1433,23 @@ extern cmdesc *CM;
 extern int TOP_CM,CM_SZ;
 extern int mclip;
 
-/* flags on menus */
+/* menu flags */
 extern
-char *cm_v_skel,
-     *cm_v_border,
+char *cm_v_small,
+     *cm_v_medium,
+     *cm_v_large,
+     *cm_v_def,
+     *cm_v_hide,
+     *cm_v_of,
      *cm_v_wclip,
      *cm_v_map,
-     *cm_v_hs,
-     *cm_v_hb,
      *cm_v_vhs,
      *cm_v_cc,
-     *cm_v_closures,
-     *cm_v_symbols,
-     *cm_v_words,
-     *cm_v_glines,
      *cm_v_mat,
      *cm_v_mat_k,
      *cm_v_cmp_k,
      *cm_v_cmp,
      *cm_v_st,
-     *cm_v_bo,
-     *cm_v_io,
 
      *cm_e_od,
      *cm_e_rescan,
@@ -1381,20 +1463,13 @@ char *cm_v_skel,
      *cm_e_sw,
      *cm_e_sh,
      *cm_e_sn,
+     *cm_e_ac,
 
      *cm_o_curr,
      *cm_o_all,
-     *cm_o_small,
-     *cm_o_medium,
-     *cm_o_large,
-     *cm_o_def,
-     *cm_o_hide,
-     *cm_o_of,
-     *cm_o_rs,
-     *cm_o_bb,
      *cm_o_dkeys,
      *cm_o_amenu,
-     *cm_o_sum,
+     *cm_o_pgo,
 
      *cm_a_arabic,
      *cm_a_cyrillic,
@@ -1404,15 +1479,36 @@ char *cm_v_skel,
      *cm_a_kana,
      *cm_a_abbrev,
      *cm_a_number,
-     *cm_a_ideogram;
+     *cm_a_ideogram,
+
+     *cm_b_skel,
+     *cm_b_border,
+     *cm_b_hs,
+     *cm_b_hb,
+
+     *cm_d_pixels,
+     *cm_d_closures,
+     *cm_d_symbols,
+     *cm_d_words,
+     *cm_d_ptype,
+     *cm_d_rs,
+     *cm_d_bb,
+
+     *cm_g_glines,
+     *cm_g_align,
+     *cm_g_lb,
+     *cm_g_bo,
+     *cm_g_io,
+     *cm_g_sum;
 
 /* input buffers */
 extern char tb[];
 
 /* OCR zone */
-#define LIMITS_SIZE 8
+#define MAX_ZONES 127
+#define LIMITS_SIZE (8*MAX_ZONES)
+extern int zones;
 extern int limits[];
-extern int limits_top;
 
 /* list of pixels of an symbol */
 typedef struct {
@@ -1506,10 +1602,12 @@ extern char *l2g[256];
 
 /*
 
-Pattern comparison result.
+Pattern comparison result and bitmap buffer.
 
 */
-extern int cp_result;
+extern int cp_result,cp_diag;
+extern unsigned cmp_bmp[];
+extern int cmp_bmp_w,cmp_bmp_h,cmp_bmp_dx,cmp_bmp_dy,cmp_bmp_x0,cmp_bmp_y0;
 
 /*
 
@@ -1521,6 +1619,13 @@ Macros used by spcpy.
 
 /*
 
+Spyhole stuff.
+
+*/
+extern int sh_tries;
+
+/*
+
 64-bit pixel mapping.
 
 */
@@ -1529,6 +1634,47 @@ extern int s64[256];
 
 /*
 
+Number of bits 1 on each 8-bit value.
+
+*/
+extern int np8[256];
+
+/*
+
+Pixel macros.
+
+The service "pixel" can be used instead of the macro "pix".
+
+*/
+#define pix(bm,bpl,i,j) \
+    (((((unsigned char *)bm)[(j)*bpl+((i)/8)]) & (1 << (7-((i)%8)))) != 0)
+#define pix3(bm,bpl,i,j) \
+    ((((((unsigned char *)bm)[(j)*bpl+((i)/8)]) >> (5-((i)%8))) & 7) == 7)
+#define setpix(bm,bpl,i,j) \
+    ((((unsigned char *)bm)[(j)*bpl+((i)/8)]) |= (1 << (7-((i)%8))))
+#define togglepix(bm,bpl,i,j) \
+    ((((unsigned char *)bm)[(j)*bpl+((i)/8)]) ^= (1 << (7-((i)%8))))
+#define unsetpix(bm,bpl,i,j) \
+    ((((unsigned char *)bm)[(j)*bpl+((i)/8)]) &= (~(1 << (7-((i)%8)))))
+
+/*
+
+This is the flea path.
+
+*/
+extern unsigned short fp[];
+extern float fsl[],fpp[];
+extern int topfp,fpsz,curr_fp,last_fp;
+
+/*
+
+Clusterization result.
+
+*/
+extern int *clusterize_r;
+
+/*
+
 Services oferred by the modules.
 
 This section must be carefully reorganized.
@@ -1552,9 +1698,12 @@ void fatal(int code,char *m, ...);
 void *c_realloc(void *p,int m,const char *s);
 void c_free(void *p);
 
-/* compressed I/O */
+/* I/O selector */
 FILE *zfopen(char *f,char *mode,int *pio);
 void zfclose(FILE *F,int pio);
+size_t zfread(void *ptr,size_t size,size_t nmemb,FILE *stream);
+size_t zfwrite(const void *ptr,size_t size,size_t nmemb,FILE *stream);
+int zfgetc(FILE *stream);
 
 /* bitmap comparison heuristics */
 int bmpcmp_map(int c,int st,int k,int direct);
@@ -1573,7 +1722,6 @@ void enter_wait(char *s,int f,int m);
 void slide(void);
 
 /* symbol pairing and word-related services */
-void summarize(int m);
 int s_pair(int a,int b,int p,int *rd);
 int rsymb(int c,int relax);
 int bsymb(int c,int relax);
@@ -1582,8 +1730,9 @@ void diag_wpairing(int t);
 void diag_pairing(int t);
 
 /* symbol services */
-int classify(int c,int bmpcmp(int,int,int,int));
+int classify(int c,int bmpcmp(int,int,int,int),int mode);
 int sdim(int k);
+int swh(int w,int h);
 int avoid(int k,int c);
 
 /* geometric services */
@@ -1591,13 +1740,17 @@ int intersize(int a,int b,int c,int d,int *e,int *f);
 float bdist(int a,int b);
 float box_dist(int a,int b,int *v);
 int ldist(int a,int b,int c,int d);
+int inside(int x,int y,int *pol,int npol);
+void comp_circ(int max);
 
 /* clipping services */
-int thisone(int x,int y);
+int symbol_at(int x,int y);
+int closure_at(int x,int y,int u);
 void list_cl(int x,int y,int w,int h,int reset);
 void list_s(int x,int y,int w,int h);
 
 /* bitmap services */
+int add_closure(cldesc *d,int dx,int dy);
 int pixel_mlist(int k);
 int wrmc8(int mm,char *s1,char *s2);
 int pixel(cldesc *c,int x,int y);
@@ -1606,11 +1759,13 @@ int byteat(int b);
 void bm2byte(char *c,unsigned char *b);
 int wrzone(char *s1,int all);
 int pbm2bm(char *f,int reset);
+int find_thing(char *p,int reset,int x,int y);
 
 /* OCR startup and steps */
 void start_ocr(int p,int s,int r);
 int build(int reset);
-int load_page(int p,int reset);
+int load_page(int p,int reset,int bin);
+int loadpgm(int reset,char *f,unsigned char **pb,int *w,int *h);
 int unload_page(int reset);
 int ocr_prep(int reset);
 
@@ -1630,6 +1785,7 @@ void mk_pattern_props(void);
 void mk_page_list(void);
 void mk_history(int n);
 void mk_page_symbol(int c);
+void mk_pattern_action(void);
 
 /* path-related services */
 void names_cpage(void);
@@ -1645,15 +1801,18 @@ void qss(int *a,int l,int r);
 void qsf(int *a,int l,int r,int inv,int cmpf(int,int));
 void qsi(int *a,int l,int r,char *t,int sz,int p,int inv);
 void qs(char *a[],int l,int r,int p,int inv);
+void true_qsi(int *a,int l,int r,char *t,int sz,int p,int inv);
 
 /* revision services */
-void review(adesc *a);
+int review(int reset,adesc *a);
 void summarize(int m);
-void from_gui(void);
+int from_gui(void);
 void synchronize(void);
 
 /* load and dump services */
 void totext(const char *fmt, ...);
+void push_text(void);
+void pop_text(void);
 int dump_session(char *f,int reset);
 int dump_acts(char *f,int reset);
 int dump_patterns(char *f,int reset);
@@ -1661,7 +1820,6 @@ int recover_session(char *f,int st,int reset);
 int recover_acts(char *f);
 int recover_patterns(char *f);
 int free_page(void);
-int load_page(int p,int reset);
 
 /* consistency services */
 int cl_cons(int c);
@@ -1670,6 +1828,7 @@ int cons(int reset);
 void consist_pg(void);
 void consist_density(void);
 void consist_magic(void);
+void consist_pp(void);
 
 /* alphabet services */
 void init_alphabet(void);
@@ -1683,6 +1842,7 @@ void cb_border(int W,int H);
 void skel(int i0,int j0,int MX,int MY);
 int skel_quality(int p);
 int tune_skel(int p);
+int tune_skel_global(int reset,int p);
 void spcpy(int to,int from);
 
 /* pattern services */
@@ -1703,6 +1863,7 @@ int compare_patterns(int p1,int p2,int bmpcmp(int,int,int,int));
 int bm2cb(unsigned char *b,int dx,int dy);
 void reset_skel_parms(void);
 void enlarge_pt(int new_toppt);
+int snbp(int s,int *bp,int *sp);
 
 /* major gui services */
 void xevents(void);
@@ -1721,6 +1882,10 @@ void get_pointer(int *x,int *y);
 void set_xfont(void);
 void init_dws(int m);
 void setview(int mode);
+void check_dlimits(int cursoron);
+void set_buttons(int s,int p);
+void symb2buttons(int s);
+void set_mclip(int f);
 
 /* auxiliar gui services */
 void copychar(void);
@@ -1728,13 +1893,14 @@ void dismiss(int f);
 void set_bl_alpha(void);
 void comp_menu_size(cmdesc *c);
 void form_auto_submit(void);
+void show_htinfo(int cx,int cy);
 
 /* merging */
 void diag_geomerge(int t);
 void geo_merge(int c);
 
 /* menu-related services */
-int additem(cmdesc *m,char *t,int p,int g);
+int additem(cmdesc *m,char *t,int p,int g,int d,char *h,int f);
 
 /* preferences stuff */
 void make_pmc(void);
@@ -1742,6 +1908,7 @@ void extr_tp(int *U,int *T,int *S,int *E,int *A,int *N,int p);
 
 /* HTML services */
 void free_earray(void);
+void img_size(char *img,int *w,int *h);
 
 /* destroy data */
 void reset();
@@ -1752,6 +1919,34 @@ int tr_align(char *a);
 int complete_align(int a,int b,int *as,int *xh,int *bl,int *ds);
 int geo_align(int c,int dd,int as,int xh,int bl,int ds);
 
+/* barcode stuff */
+int closure_border_slines(int e,int t,int crit,float val,short *res,int mr,int bar);
+int closure_border_path(int t);
+int border_path(unsigned char *b,int w,int h,unsigned short *bp,int m,int u0,int v0,int relax);
+int dist_bar(int i,int j);
+int search_barcode(void);
+int isbar(int k,float *sk,float *bl);
+float s2a(float slxy,float sly,float slx,short *bp,int i,int j);
+
+/* blockfinders */
+void cf_block(void);
+int blockfind(int reset,int pc);
+
+/* clusterization */
+int clusterize(int N,int T,int dist(int,int));
+int test_clusterize(void);
+
+/* preprocessors */
+int deskew(int reset);
+int pp_thresh(void);
+void balance(void);
+void avoid_links(float C);
+void hqbin(void);
+void test(void);
+
+/* tests */
+void internal_tests(void);
+
 /* other */
 void checkcl(int m);
 void gen_wrf();
@@ -1759,15 +1954,14 @@ void process_cl(int argc,char *argv[]);
 void process_webdata(void);
 int pagenbr(char *p);
 int s2p(int k);
-void check_dlimits(int cursoron);
-void show_htinfo(int cx,int cy);
-void set_buttons(int s,int p);
-void symb2buttons(int s);
-void img_size(char *img,int *w,int *h);
-void set_mclip(int f);
 int tsymb(int c,int relax);
 int classify_tr(char *t);
 void new_alrm(unsigned d,int who);
 void dump_cb(void);
 void init_pt(ptdesc *p);
-void cf_block(void);
+int dx(int k,int i0,int j0);
+int cma(int it,int check);
+int spyhole(int x,int y,int op,int et);
+int justify_bitmap(unsigned char *bm,int bw,int bh);
+void get_ap(int c,int *as,int *xh,int *bl,int *ds);
+void cb2bm(unsigned char *b);
