@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 1999-2001 Ricardo Ueda Karpischek
+  Copyright (C) 1999-2002 Ricardo Ueda Karpischek
 
   This is free software; you can redistribute it and/or modify
   it under the terms of the version 2 of the GNU General Public
@@ -73,11 +73,15 @@ void enlarge_pt(int new_toppt)
 
         toppt = new_toppt;
     }
+
+    else
+        db("enlarge_pt called with invalid parameter %d",new_toppt);
 }
 
 /*
 
-Copy the bitmap b to the buffer cb. Returns the number of black pixels.
+Copy the FSxFS bitmap b to the buffer cb. Returns the number of
+black pixels.
 
 */
 int bm2cb(unsigned char *b,int dx,int dy)
@@ -106,7 +110,7 @@ int bm2cb(unsigned char *b,int dx,int dy)
 
 /*
 
-Copy the buffer cb to the bitmap b.
+Copy the buffer cb to the FSxFS bitmap b.
 
 */
 void cb2bm(unsigned char *b)
@@ -139,7 +143,6 @@ Compute the number of black pixels of the pattern s.
 int snbp(int s,int *bp,int *sp)
 {
     static int a=0;
-    static int np[256];
     int i,j,n,m;
     pdesc *d;
     unsigned char *p;
@@ -148,32 +151,39 @@ int snbp(int s,int *bp,int *sp)
     if (a==0) {
         for (i=0; i<256; ++i) {
             n = i;
-            np[i] = 0;
+            np8[i] = 0;
             for (j=0; j<8; ++j) {
-                np[i] += (n&1);
+                np8[i] += (n&1);
                 n >>= 1;
             }
         }
         a = 1;
     }
 
+    /*
+        If the caller informed an invalid pattern index, return (that
+        may happen to just initialize np8).
+    */
+    if (s < 0)
+        return(0);
+
     /* compute the number of pixels of the pattern s */
     d = pattern + s;
     for (p=(unsigned char *)d->b, i=n=0; i<FS*(FS/8); ++p, ++i)
-        n += np[*p];
+        n += np8[*p];
     if (bp != NULL)
         *bp = n;
     if (sp != NULL) {
         for (p=(unsigned char *)d->s, i=m=0; i<FS*(FS/8); ++p, ++i)
-            m += np[*p];
-            *sp = m;
+            m += np8[*p];
+        *sp = m;
     }
     return(n);
 }
 
 /*
 
-Compute the skeleton of the pattern p.
+Compute the skeleton of the pattern c.
 
 */
 void pskel(int c)
@@ -324,6 +334,76 @@ void justify_pattern(short *h,short *w,short *dx,short *dy)
 
 /*
 
+Copy the bitmap on buffer bm to bmp_cmp, performing top- and
+left-justification. Fails if the width or height are larger than
+FS (in this case returns 0). Returns 1 when successfull.
+
+*/
+int justify_bitmap(unsigned char *bm,int bw,int bh)
+{
+    int b,i,j,bpl;
+
+    bpl = (bw/8) + ((bw%8)!=0);
+
+    /* compute left margin */
+    b = 0;
+    for (i=0; (b==0) && (i<bw); ++i) {
+        for (j=0; (b==0) && (j<bh); ++j)
+            b = pix(bm,bpl,i,j);
+    }
+    if (b == 0)
+        i = 0;
+    else
+        --i;
+    cmp_bmp_dx = i;
+
+    /* compute width */
+    b = 0;
+    for (cmp_bmp_w=bw; (b==0) && (--cmp_bmp_w>=0);) {
+        for (j=0; (b==0) && (j<bh); ++j)
+            b = pix(bm,bpl,cmp_bmp_w,j);
+    }
+    cmp_bmp_w -= (cmp_bmp_dx-1);
+    if (cmp_bmp_w > FS)
+        return(0);
+
+    /* compute top margin */
+    b = 0;
+    for (j=0; (b==0) && (j<bh); ++j) {
+        for (i=0; (b==0) && (i<bw); ++i)
+            b = pix(bm,bpl,i,j);
+    }
+    if (b == 0)
+        j = 0;
+    else
+        --j;
+    cmp_bmp_dy = j;
+
+    /* compute height */
+    b = 0;
+    for (cmp_bmp_h=bh; (b==0) && (--cmp_bmp_h>=0);) {
+        for (i=0; (b==0) && (i<bw); ++i)
+            b = pix(bm,bpl,i,cmp_bmp_h);
+    }
+    cmp_bmp_h -= (cmp_bmp_dy-1);
+    if (cmp_bmp_h > FS)
+        return(0);
+
+    /* copy to cmp_bmp (non-optimized) */
+    memset(cmp_bmp,0,BMS);
+    for (i=0; i<cmp_bmp_w; ++i) {
+        for (j=0; j<cmp_bmp_h; ++j) {
+            if (pix(bm,bpl,i+cmp_bmp_dx,j+cmp_bmp_dy))
+                setpix(cmp_bmp,BLS,i,j);
+        }
+    }
+
+    /* success */
+    return(1);
+}
+
+/*
+
 Alloc new pattern.
 
 */
@@ -465,8 +545,11 @@ void update_pattern_raw(int new,int an,int mc,char *tb,char a,short ft,short fs,
     if (new) {
         int bp,sp;
 
+        /*
         if (mc >= 0)
             d->e = mc;
+        */
+
         d->ds = strlen(pagename) + 1;
         d->d = c_realloc(NULL,d->ds,NULL);
         strcpy(d->d,pagename);
@@ -508,6 +591,7 @@ void update_pattern_raw(int new,int an,int mc,char *tb,char a,short ft,short fs,
         d->id = ++s_id;
         ++topp;
         snprintf(mba,MMB,"pattern %d added to font",topp);
+        tskel_ready = 0;
         show_hint(1,mba);
     }
 
@@ -526,6 +610,7 @@ to the pattern array.
 BUG: expensive search.
 
 */
+/*
 int s2p(int k)
 {
     int c,j;
@@ -539,11 +624,12 @@ int s2p(int k)
     }
     return(c);
 }
+*/
 
 /*
 
 Add or update one pattern. The bitmap is obtained from the symbol
-k of the current page. If k<0, then updates the cdfc entry.
+k of the current page. If k<0, then update the cdfc entry.
 
 This function is just a higher-level interface for the
 update_pattern function. It obtains the current location of the
@@ -566,7 +652,12 @@ int update_pattern(int k,char *tr,int an,char a,short pt,short fs,int f)
     */
     if (k >= 0) {
 
+        /* the link between pattern and symbol does not exist anymore */
+        /*
         cdfc = s2p(k);
+        */
+
+        cdfc = -1;
 
         if (cdfc < 0) {
 
@@ -575,12 +666,19 @@ int update_pattern(int k,char *tr,int an,char a,short pt,short fs,int f)
         }
     }
 
+    /*
+        Prepare the spyhole buffer for inclusion.
+    */
+    else if (cdfc < 0) {
+        spyhole(0,0,2,0);
+    }
+
     /* add or update pattern */
-    if ((k >= 0) && (cdfc < 0)) {
+    if (cdfc < 0) {
         update_pattern_raw(1,an,k,tr,a,pt,fs,f);
         cdfc = topp;
     }
-    
+
     /* update pattern */
     else if (cdfc >= 0)
         update_pattern_raw(0,an,k,tr,a,pt,fs,f);
@@ -688,6 +786,7 @@ void rm_pattern(int n)
     --topp;
 
     font_changed = 1;
+    tskel_ready = 0;
 
     /* must regenerate the pattern-related windows */
     dw[PATTERN].rg = 1;
@@ -781,7 +880,10 @@ int prepare_patterns(int reset)
     if (reset) {
         c = 0;
         st = 1;
-        show_hint(1,"tuning skels");
+        show_hint(3,NULL);
+        show_hint(0,"tuning skels");
+        if (!tskel_ready)
+            tune_skel_global(1,-1);
         return(1);
     }
 
@@ -793,30 +895,51 @@ int prepare_patterns(int reset)
 
         /* finished */
         if (c > topp) {
-            show_hint(1,"recognizing..");
+            show_hint(0,"computing skeletons..");
             st = 2;
             c = 0;
             return(1);
         }
 
         /* compute skeleton of pattern c */
-        else if ((st_auto) && (pattern[c].p[0] < 0)) {
+        else if (st_auto && (!tskel_ready)) {
+            int r;
+
+            show_hint(0,"auto-tuning now at %d/%d",c,topp+1);
 
             /* invalidate current skeleton */
             pattern[c].sw = -1;
 
-            if (*cm_v_st != ' ') {
-                if (!dw[TUNE_SKEL].v) {
-                    setview(TUNE_SKEL);
-                }
+            /* accumulate per-candidate qualities for pattern c */
+            r = tune_skel_global(0,c);
+
+            if (r == 0) {
+                spcpy(SP_DEF,SP_GLOBAL);
             }
 
-            tune_skel(c);
+            /*
+                display the skeleton, if requested
 
-            font_changed = 1;
+                OOPS.. each pattern is displayed once (using its
+                best parameters), not once per candidate.
+            */
+            if ((r==0) || (*cm_v_st != ' ')) {
+                if ((*cm_v_st != ' ') && (!dw[TUNE_SKEL].v)) {
+                    setview(TUNE_SKEL);
+                }
+                cdfc = c;
+                dw[TUNE_SKEL].rg = 1;
+                dw[TUNE_PATTERN].rg = 1;
+                redraw_dw = 1;
+            }
+
+            /* prepare the next */
+            if (++c > topp)
+                tskel_ready = 1;
         }
 
-        ++c;
+        else
+            c = topp+1;
     }
 
     /*
@@ -827,7 +950,7 @@ int prepare_patterns(int reset)
 
         /* finished */
         if (c > topp) {
-            show_hint(1,"recognizing..");
+            show_hint(0,"distributing into types..");
             st = 3;
             c = 0;
             return(1);
@@ -850,8 +973,12 @@ int prepare_patterns(int reset)
         int k;
         pdesc *d;
 
+        /*
+            NOT READY. WHEN READY, CHANGE k<=-1 to k<=topp
+        */
+
         /* compute the pattern types */
-        for (k=0; k<=topp; ++k) {
+        for (k=0; k<=-1; ++k) {
             d = pattern + k;
 
             if ((d->pt < 0) && (tr_align(d->tr) == 12)) {
@@ -900,12 +1027,13 @@ int prepare_patterns(int reset)
         st = 4;
         i = 0;
         j = 0;
+        show_hint(0,"searching false positives..");
         return(1);
     }
 
     /*
         ** STATE 4 **
-        search false matches
+        search false positives
     */
     else if (st == 4) {
 
@@ -917,14 +1045,27 @@ int prepare_patterns(int reset)
                 i = 0;
                 j = 0;
                 st = 5;
+                show_hint(0,"searching unexpected mismatches..");
                 return(1);
             }
             else
                 j = i+1;
         }
 
+        /*
         compare_patterns(-1,0,bmpcmp_skel);
         while (compare_patterns(i,j,bmpcmp_skel));
+        */
+
+        {
+            int t;
+
+            classify(-1,NULL,0);
+            t = this_pattern;
+            this_pattern = j;
+            while (classify(i,bmpcmp_skel,2));
+            this_pattern = t;
+        }
 
         /*
         printf("patterns %d and %d: %d\n",i,j,cp_result);
@@ -938,8 +1079,8 @@ int prepare_patterns(int reset)
     else if (st == 5) {
         pdesc *p,*q;
 
-        if (*cm_o_sum == ' ') {
-            j = i = topp;
+        if (*cm_g_sum == ' ') {
+            i = topp;
         }
 
         if (++j > topp) {
@@ -947,6 +1088,7 @@ int prepare_patterns(int reset)
                 i = 0;
                 j = 0;
                 st = 0;
+                show_hint(3,NULL);
                 return(0);
             }
             else
@@ -960,12 +1102,23 @@ int prepare_patterns(int reset)
             (p->tr != NULL) && (q->tr != NULL) &&
             (strcmp(p->tr,q->tr) == 0)) {
 
+            int t;
+
+            /*
             compare_patterns(-1,0,bmpcmp_skel);
             while (compare_patterns(i,j,bmpcmp_skel));
+            */
+
+            classify(-1,NULL,0);
+            t = this_pattern;
+            this_pattern = j;
+            while (classify(i,bmpcmp_skel,2));
+            this_pattern = t;
         }
 
         /*
-        printf("patterns %d and %d: %d\n",i,j,cp_result);
+        if (cp_result >= 0)
+            printf("patterns %d and %d: %d\n",i,j,cp_result);
         */
     }
 
