@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 1999-2001 Ricardo Ueda Karpischek
+  Copyright (C) 1999-2002 Ricardo Ueda Karpischek
 
   This is free software; you can redistribute it and/or modify
   it under the terms of the version 2 of the GNU General Public
@@ -41,7 +41,7 @@ full, its size is increased by AB_INCR.
 
 /*
 
-Total memory allocated to bitmaps (to avoid crashes).
+Total memory allocated for bitmaps (to avoid crashes).
 
 */
 int ab_mem;
@@ -49,7 +49,7 @@ int ab_mem;
 /*
 
 To avoid crashes, we limit the maximum amount of memory to be
-allocated to bitmaps. The segmentation algorithm will alloc too
+allocated for bitmaps. The segmentation algorithm will alloc too
 much memory when the page is too dark. This is not a "bug", but a
 "naive method".
 
@@ -76,19 +76,20 @@ remaining bits (if any) are left blank, and so on. The leftmost
 bit on each byte is the most significative one (black, or "on",
 is 1, and white, or "off" is 0). An example follows:
 
+       76543210765432
       +--------------+
-      |              |
-      |   XX XXXX    |
-      |    XX   XX   |
-      |    XX   XX   |
-      |    XX   XX   |
-      |    XX   XX   |
-      |              |
+      |              | 00000000 00000000 =  0   0
+      |   XX XXXX    | 00011011 11000000 = 27 192
+      |    XX   XX   | 00001100 01100000 = 12  96
+      |    XX   XX   | 00001100 01100000 = 12  96
+      |    XX   XX   | 00001100 01100000 = 12  96
+      |    XX   XX   | 00001100 01100000 = 12  96
+      |              | 00000000 00000000 =  0   0
       +--------------+
 
       stored as: 0 0 27 192 12 96 12 96 12 96 12 96 0 0
 
-Note that the array of bytes that encodesone bitmap does not
+Note that the array of bytes that encodes one bitmap does not
 contain the bitmap width nor the height. So bitmaps must be
 stored together with other data. This is done by structures where
 the bitmap is one field and the geometric information is stored
@@ -107,11 +108,11 @@ temporary storage to build the closures while reading each line
 of the scanned document. The function pbm2bm alloc a temporary
 array of these structures (the ab array).
 
-The fields x and y of the bdesc structure store the leftmost
-column and topmost line occuped by the bitmap within the
-document. The fields w and h store the width and heigth of the
-bitmap. The field s stores the size of the memory buffer point by
-the field b (used to store the bitmap itself).
+The fields x and y of the bdesc structure store the bitmap
+leftmost column and topmost line (document-relative
+coordinates). The fields w and h store the width and heigth of
+the bitmap. The field s stores the size of the memory buffer
+point by the field b (used to store the bitmap itself).
 
 When x<0, the bdesc strucuture is not being used, but the field b
 may be pointing to a valid memory buffer. In this case, the field
@@ -171,9 +172,9 @@ int byteat(int b)
 Dump bitmap to asc.
 
 The size of the memory area pointed by s must be at least
-b->h * (b->l+1) + 1.
+b->h * (b->w+1) + 1.
 
-This service is just a debugging facility.
+This service is a debugging facility.
 
 */
 void bdump(char *s,bdesc *b)
@@ -263,9 +264,14 @@ void extend(bdesc *b,int x,int y,int w,int h)
 
     /* enlarge buffer */
     if (b->s < bl*h) {
-        int ns;
+        int a,ns;
 
+        /* large bitmaps probably will grow more, so enlarge it by at least 10% */
         ns = bl * (h+10);
+        if (ns < (a=b->s+b->s/10))
+            ns = a;
+
+        /* enlarge */
         ab_mem += ns - b->s;
         test_ab_mem();
         b->s = ns;
@@ -501,14 +507,31 @@ void bitmerge(bdesc *b,bdesc *c)
     h = y2-y1+1;
 
     /* extend b */
-    if ((w>b->w) || (h>b->h))
+    if ((w>b->w) || (h>b->h)) {
+
+        /*
+            large bitmaps probably will continue growing, so add
+            FS bits on each side.
+        */
+        if (w > 5*FS) {
+            if (x1 > FS)
+                x1 -= FS;
+            else
+                x1 = 0;
+            if (x2+=FS >= XRES)
+                x2 = XRES-1;
+            w = x2-x1+1;
+        }
+
+        /* now extend */
         extend(b,x1,y1,w,h);
+    }
 
     /* bytes per line */
     bl = byteat(b->w);
     pbl = byteat(c->x-x1);
 
-    /* copy bits from c to the extended bitmap */
+    /* copy c to the extended bitmap */
     n = 128;
     q = c->b;
     m0 = ((unsigned) 1) << (7-((c->x-x1)%8));
@@ -543,7 +566,7 @@ void bitmerge(bdesc *b,bdesc *c)
 /*
 
 Make *p and *q be the same active bitmap index, merging
-*q to *p if necessary.
+*q to *p (or *p to *q) if necessary.
 
 This function exists only to simplify a loop on pbm2bm (take a
 look on the calls).
@@ -561,13 +584,35 @@ void join(int *p,int *q,bdesc *ab) {
     if (*q >= 0) {
         if ((*p >= 0) && (*p != *q)) {
             if (ab[*q].x >= 0) {
+                int old;
 
-                bitmerge(ab+*p,ab+*q);
+                /*
+                    merge the smaller to the larger.
+                */
+                if (ab[*p].s >= ab[*q].s) {
+                    bitmerge(ab+*p,ab+*q);
+                    ab[*q].x = -1;
+                    ab[*q].y = *p;
+                    old = *q;
+                    *q = *p;
+                }
+                else {
+                    bitmerge(ab+*q,ab+*p);
+                    ab[*p].x = -1;
+                    ab[*p].y = *q;
+                    old = *p;
+                    *p = *q;
+                }
 
-                ab[*q].x = -1;
-                ab[*q].y = *p;
+                /* free large unused memory buffers */
+                if (ab[old].s > 64000) {
+                    c_free(ab[old].b);
+                    ab[old].b = NULL;
+                    ab[old].s = 0;
+                }
             }
-            *q =*p;
+            else
+                *q = *p;
         }
         else
             *p = *q;
@@ -578,8 +623,10 @@ void join(int *p,int *q,bdesc *ab) {
 
 Create a new closure and the associated symbol.
 
+(this service was obsoleted by new_cl)
+
 */
-void new_cl(int x,int y,int w,int h,void *b)
+void new_cl_global(int x,int y,int w,int h,void *b)
 {
     int t;
     unsigned char m;
@@ -624,11 +671,208 @@ void new_cl(int x,int y,int w,int h,void *b)
 
 /*
 
+Create a new closure and the associated symbol.
+
+*/
+void new_cl(int x,int y,int w,int h,void *b)
+{
+    int t,ct;
+    unsigned char m,n,*p,*q;
+    int i,j;
+    int tm,lm,bm,rm,cw,ch;
+
+    /* must enlarge allocated area for cl */
+    checkcl(++topcl);
+
+    /*
+        Compute the topmost, bottommost, leftmost and rightmost
+        coordinates, number of black pixels and locate a seed.
+    */
+    cl[topcl].nbp = 0;
+    t = byteat(w);
+    tm = h;
+    bm = -1;
+    lm = w;
+    rm = -1;
+    for (j=0; j<h; ++j) {
+        m = 128;
+        p = ((unsigned char *)b) + j*t;
+        for (i=0; i<w; ++i) {
+            if (*p & m) {
+
+                /* adjust limits as required */
+                if (i < lm)
+                    lm = i;
+                if (i > rm)
+                    rm = i;
+                if (j < tm)
+                    tm = j;
+                if (j > bm)
+                    bm = j;
+
+                /* take (i,j) as seed if it's the first black pixel found */
+                if (++(cl[topcl].nbp) == 1) {
+                    cl[topcl].seed[0] = x+i;
+                    cl[topcl].seed[1] = y+j;
+                }
+            }
+            if ((m >>= 1) == 0) {
+                ++p;
+                m = 128;
+            }
+        }
+    }
+
+    /* computed geometry */
+    cw = rm-lm+1;
+    ch = bm-tm+1;
+
+    /* alloc memory buffer to the closure bitmap */
+    ct = byteat(cw);
+    cl[topcl].bm = c_realloc(NULL,ct*ch,"new_cl");
+
+    /*
+        In some cases, the bitmap must be copied line
+        by line.
+    */
+    if ((lm > 0) || (tm > 0) || (cw < w) || (ch < h)) {
+        memset(cl[topcl].bm,0,ct*ch);
+        for (j=0; j<ch; ++j) {
+            p = ((unsigned char *)b) + (j+tm)*t + (lm/8);
+            m = 1 << (7-(lm%8));
+            q = ((unsigned char *)cl[topcl].bm) + j*ct;
+            n = 128;
+            for (i=0; i<cw; ++i) {
+                if (*p & m) {
+                    *q |= n;
+                }
+                if ((m >>= 1) == 0) {
+                    ++p;
+                    m = 128;
+                }
+                if ((n >>= 1) == 0) {
+                    ++q;
+                    n = 128;
+                }
+            }
+        }
+    }
+
+    /* simple case: copy the bitmap using memcpy */
+    else {
+        memcpy(cl[topcl].bm,b,ct*ch);
+    }
+
+    /* initialize the new entry */
+    cl[topcl].l = x + lm;
+    cl[topcl].r = x + rm;
+    cl[topcl].t = y + tm;
+    cl[topcl].b = y + bm;
+    cl[topcl].sup = c_realloc(NULL,MAXSUP*sizeof(int),NULL);
+    cl[topcl].sup[0] = -1;
+    cl[topcl].supsz = MAXSUP;
+
+    /* create unitary symbol */
+    C_SET(mc[new_mc(&topcl,1)].f,F_ISP);
+}
+
+/*
+
+Search next interior pixel and apply the spyhole to extract the
+closure.
+
+*/
+int find_thing(char *p,int reset,int x,int y)
+{
+    static int i,j;
+    int u,v,k,l,f,r;
+
+    if (reset) {
+        i = j = 0;
+        return(1);
+    }
+
+    if (x < 0) {
+        u = i;
+        v = j;
+    }
+    else {
+        u = x;
+        v = y;
+    }
+
+    /* try to find and extract one closure */
+    for (f=0; (f==0) && (v<YRES); ) {
+
+        /* (u,v) is not black enough */
+        r = (pixmap[u+v*XRES] > thresh_val);
+
+        if (!r) {
+
+            /* got one? yes! */
+            if (spyhole(u,v,4,thresh_val)) {
+
+                /* add as new closure */
+                spyhole(0,0,3,0);
+                new_cl(cmp_bmp_x0,cmp_bmp_y0,FS,FS,cmp_bmp);
+	
+                /* remove from the pixmap */
+                spyhole(0,0,1,0);
+                for (k=0; k<FS; ++k) {
+                    for (l=0; l<FS; ++l) {
+                        if (pix(cmp_bmp,BLS,k,l)) {
+                            pixmap[(l+cmp_bmp_y0)*XRES+k+cmp_bmp_x0] = 255;
+                        }
+                    }
+                }
+
+                /* get out from the loop */
+                f = 1;
+            }
+	
+            /*
+                spyhole() disagreed with pbm2bm(). So one of them
+                is probably buggy.
+            */
+            else {
+                db("hey, what's running on at %d %d?\n",u,v);
+            }
+        }
+
+        /* clear it */
+        else {
+            pixmap[u+v*XRES] = 255;
+        }
+
+        /* prepare the next position */
+        if (++u >= XRES) {
+            ++v;
+            u = 0;
+        }
+
+        /*
+            the caller is supposed to inform a true closure seed.
+        */
+        if ((!f) && (x>0)) {
+            db("this caller got drunk!");
+        }
+    }
+
+    show_hint(1,"now at line %d, total %d closures",j,topcl+1);
+
+    /* remember position */
+    i = u;
+    j = v;
+
+    return(f);
+}
+
+/*
+
 Read pbm file and convert its closures to bitmaps.
 
-Some folks may think this function is a bug because it makes very
-hard to add support to other graphic formats. Yes, it's
-true. Someday I'll split it in two separate functions.
+Some folks may think this function is a stupid thing because it
+makes very hard to add support to other graphic formats.
 
 Anyway, it's design has two good properties. The first is that it
 reads the bitmap only once, so one could run a format conversion
@@ -665,6 +909,9 @@ int pbm2bm(char *f,int reset)
     */
     static int *kl,*ll;
 
+    /* account pixels eaten by neighbours */
+    static int eaten;
+
     if (reset) {
 
         int c,rc;
@@ -685,7 +932,7 @@ int pbm2bm(char *f,int reset)
         F = zfopen(f,"r",&pio);
 
         /* P4 magic number */
-        if ((fgetc(F) != 'P') || (fgetc(F) != '4')) {
+        if ((zfgetc(F) != 'P') || (zfgetc(F) != '4')) {
             fatal(DI,"%s is not a pbm raw file",f);
         }
 
@@ -693,7 +940,7 @@ int pbm2bm(char *f,int reset)
             Reading loop. PBM defines "whitespace" as any from space,
             tab, CR or LF. Comments start with '#' and end on LF.
         */
-        for (n=rc=0, w=h=0; (n<4) && ((c=fgetc(F)) != EOF); ) {
+        for (n=rc=0, w=h=0; (n<4) && ((c=zfgetc(F)) != EOF); ) {
 
             /* reading comment */
             if ((c == '#') || ((rc==1) && (c!='\n'))) {
@@ -761,6 +1008,7 @@ int pbm2bm(char *f,int reset)
 
         /* prepare reading loop and return */
         n = 0;
+        eaten = 0;
         return(1);
     }
 
@@ -780,7 +1028,7 @@ int pbm2bm(char *f,int reset)
 
         /* read one line from the document */
         if (n < h) {
-            if (fread(s,1,bl,F) < bl) {
+            if (zfread(s,1,bl,F) < bl) {
                 fatal(IO,"error reading file \"%s\"",f);
             }
         }
@@ -890,7 +1138,82 @@ int pbm2bm(char *f,int reset)
                 ab[j].y = -ab[j].y - 1;
 
                 /* create the closure */
-                new_cl(ab[j].x,ab[j].y,ab[j].w,ab[j].h,ab[j].b);
+                if (localbin &&
+                    (ab[j].w < FS) &&
+                    (ab[j].h < FS) &&
+                    (ab[j].x > FS) &&
+                    ((XRES-ab[j].x) > FS) &&
+                    (ab[j].y > FS) &&
+                    ((YRES-ab[j].y) > FS)) {
+
+                    int f,u,v,t,m;
+                    unsigned char *p;
+
+                    f = 0;
+
+                    /*
+                        search a bitmap seed and call find_thing.
+                    */
+                    t = byteat(ab[j].w);
+                    for (v=0; (f==0) && (v<ab[j].h); ++v) {
+                        m = 128;
+                        p = ((unsigned char *)ab[j].b) + v*t;
+                        for (u=0; (f==0) && (u<ab[j].w); ++u) {
+                            if (*p & m) {
+                                int x,y;
+
+                                x = u+ab[j].x;
+                                y = v+ab[j].y;
+                                if (pixmap[x+y*XRES] <= thresh_val)
+                                    find_thing(NULL,0,x,y);
+                                else {
+
+                                    /*
+                                        The pixel (x,y) is set in ab[j], but
+                                        cleared in pixmap. The reason is: one
+                                        bitmap from the line y-1 grow using
+                                        a high threshold and ate (x,y). Just
+                                        account.
+                                    */
+                                    eaten++;
+                                }
+                                f = 1;
+                            }
+                            else if ((m >>= 1) == 0) {
+                                ++p;
+                                m = 128;
+                            }
+                        }
+                    }
+
+                    /*
+                        Search (globally) a bitmap seed and call find_thing.
+                        This code is buggy because searching the entire page
+                        may get as seed a pixel from another symbol. But the
+                        code is being maintained for eventual tests.
+                    */
+                    /*
+                    for (u=ab[j].w; (f==0) && (u>0); --u) {
+                        for (v=ab[j].h; (f==0) && (v>0); --v) {
+                            if (pixmap[(u+ab[j].x)+(v+ab[j].y)*XRES] < thresh_val) {
+                                find_thing(NULL,0,u+ab[j].x,v+ab[j].y);
+                                f = 1;
+                            }
+                        }
+                    }
+                    */
+                }
+
+                /*
+                    BUG: when the page contains symbols close to the
+                    page boundaries, new_cl is called directly even if
+                    local binarizations was chosen. So simple
+                    thresholding will be applied and the pixels won't
+                    be removed from the pixmap.
+                */
+                else {
+                    new_cl(ab[j].x,ab[j].y,ab[j].w,ab[j].h,ab[j].b);
+                }
 
                 /* mark ab[j] as unused */
                 ab[j].x = -1;
@@ -987,7 +1310,7 @@ int pbm2bm(char *f,int reset)
 
 /*
 
-Write the current zone as a raw pbm file.
+Write the first zone as a raw pbm or pgm file.
 
 */
 int wrzone(char *s1,int all)
@@ -1009,20 +1332,43 @@ int wrzone(char *s1,int all)
         }
     }
 
-    /* mc limits */
+    /* compute the smallest rectangle that contains the first zone */
     l = limits[0];
+    if (l > limits[2])
+        l = limits[2];
     r = limits[6];
+    if (r < limits[4])
+        r = limits[4];
     t = limits[1];
+    if (t > limits[7])
+        r = limits[7];
     b = limits[5];
+    if (b < limits[3])
+        b = limits[3];
     w = r-l+1;
     h = b-t+1;
 
+    /* save as PGM */
+    if ((pixmap != NULL) && (pm_t == 8)) {
+        int F,j;
+
+        F = open(s1,O_WRONLY|O_CREAT|O_TRUNC,0644);
+        sprintf((char *)g,"P5\n%d %d\n255\n",w,h);
+        write(F,g,strlen((char *)g));
+        for (j=0; j<h; ++j) { 
+            write(F,pixmap+(t+j)*XRES+l,w);
+        }
+        close(F);
+        show_hint(2,"%s successfully written",s1);
+
+        return(1);
+    }
+
     /* alloc memory */
     RWX = r-l+1;
     RWY = b-t+1;
     bpl = (RWX/8) + ((RWX%8) != 0);
-    if ((rw = alloca(RWY*bpl)) == NULL)
-        return(0);
+    rw = c_realloc(NULL,RWY*bpl,NULL);
     memset(rw,0,RWY*bpl);
 
     /* look for closures that intersect the rectangle (l,r,t,b) */
@@ -1097,6 +1443,7 @@ int wrzone(char *s1,int all)
         show_hint(2,"%s successfully written",s1);
     }
 
+    c_free(rw);
     return(1);
 }
 
